/* --------------------------------------------------------------  */
/* (C)Copyright 2008                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef __MC_RAND_SB_SEEDS__
#define __MC_RAND_SB_SEEDS__ 1

int mc_rand_sb_distribute_sims(int n_simulation, int block_size, int n_spu, int *n_spu_sim);
int mc_rand_sb_create_spu_seeds(unsigned int u32Seed, int n_rand_per_sim, int n_spu, int *n_spu_sim, unsigned int *spu_seed);

//  Function mc_rand_sb_distribute_sims
//
//	Input parameters:
//	    int n_simulation:  the total number of simulations requested by the calling program
//			       n_simulation must be greater than zero
//	    int block_size:    the size of the blocks of simulations that will be allocated to each SPU
//			       block_size must be at least four and an integer power of two
//	    int n_spu:         the number of SPUs that will run the simulations
//			       n_spu must be greater than zero
//	Output parameter:
// 	    int *n_spu_sim:    an array containing the number of simulations that each SPU will run
//			       the array must have n_spu elements
//	Return value:
//	    Parameter error:   -1
//	    Otherwise:         the actual number of simulations being run, which is the requested
//			       number of simulations rounded up to the next whole block

int mc_rand_sb_distribute_sims(int n_simulation, int block_size, int n_spu, int *n_spu_sim) {
    int j, k, *np;

    /* Check that the number of simulations is greater than zero */
    if(n_simulation <= 0) return(-1);

    /* Check that the block size is at least four */
    if(block_size <= 4) return(-1);

    /* Find the log to the base 2 of the block size */
    int log2_block_size = 0;
    k = block_size >> 1;
    while(k > 0) {
	k = k >> 1;
	++log2_block_size;
    }

    /* Check that the block size is a power of two */
    if(block_size != (1 << log2_block_size)) return(-1);

    /* Check that the number of SPUs is greater than zero */
    if(n_spu <= 0) return(-1);

    int n_block = (n_simulation + block_size - 1) >> log2_block_size;
    int n_spu_block = n_block / n_spu;
    int n_rem_block = n_block - (n_spu_block * n_spu);
    int n_sim = 0;

    /* These SPUs get an extra block */
    int n_spu_simulation = (n_spu_block << log2_block_size) + block_size;
    for(j = 0, np = n_spu_sim; j < n_rem_block; ++j, ++np) {
    	*np = n_spu_simulation;
	n_sim += n_spu_simulation;
    }

    /* These SPUs do not */
    n_spu_simulation -= block_size;
    for( ; j < n_spu; ++j, ++np) {
    	*np = n_spu_simulation;
	n_sim += n_spu_simulation;
    }

    /* Return the number of simulations rounded up the the next whole block */
    return(n_sim);
}

//  Function mc_rand_sb_create_spu_seeds
//
//	Include file:
//	    sobol_init_30_40.h
//	Input parameters:
//	    unsigned int u32Seed: the seed for the Sobol generator of the original serial code
//	    int n_rand_per_sim:   the number of random numbers that each simulations requires
//				  n_rand_per_sim must be greater than zero
//	    int n_spu:            the number of SPUs that will run the simulations
//			          n_spu must be greater than zero
// 	    int *n_spu_sim:       an array containing the number of simulations that each SPU will run
//			          the array must have n_spu elements
//				  the number of simulations must be greater than zero and a multiple of four
//	Output parameter:
//	    unsigned int *spu_seed: an array containing the seed for the Sobol generator for each SPU
//			            the array must have n_spu elements
//	Return value:
//	    Parameter error:	  -1
//	    Otherwise:            the total number of simulations being run

int mc_rand_sb_create_spu_seeds(unsigned int u32Seed, int n_rand_per_sim, int n_spu, int *n_spu_sim, unsigned int *spu_seed) {
    int j, *np;
    unsigned int *sp;

    /* Check that the number of random numbers per simulation is greater than zero */
    if(n_rand_per_sim <= 0) return(-1);

    /* Check that the number of SPUs is greater than zero */
    if(n_spu <= 0) return(-1);

    int n_sim = 0;
    for(j = 0, np = n_spu_sim, sp = spu_seed; j < n_spu; ++j, ++np, ++sp) {

	/* Check that the number of simulations for each SPU is greater than zero */
	int n_spu_simulation = *np;
	if(n_spu_simulation <= 0) return(-1);

	/* Check that the number of simulations for each SPU is a multiple of four */
	if((n_spu_simulation & 3) != 0) return(-1);

	*sp = (n_rand_per_sim * (n_sim >> 2)) + u32Seed;
	n_sim += n_spu_simulation;
    }

    return(n_sim);
}

#endif  /* __MC_RAND_SB_SEEDS__ */
