/* --------------------------------------------------------------  */
/* (C)Copyright 2008                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

// piMC_ppu.c
//   Monte Carlo simulation to calculate PI in parallel on several
//   spe processors.  This example shows the usage of the ppu apis
//   mc_rand_sb_distribute_sims and mc_rand_sb_create_spu_seeds.
//   When given the number of simulations to be run and the number
//   of spes to be used, these apis create appropriate values for 
//   use in the Sobol spe seeding function mc_rand_sb_seed.

#include <libspe2.h>
#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc_align.h>
#include <free_align.h>

// NOTE: The following three includes have an order dependency to ensure that Sobol control block is defined.
#include <mc_rand_sb.h>
#include "../piMC.h"
#include <stdint.h>
#include "../../sobol_init_30_40.h"
#include "../../mc_rand_sb_seeds.h"

#define NUMSPUS 16 

extern spe_program_handle_t calcPI_spu;

// Static allocation of memory aligned to 128 byte boundary for the control block.
// A control block for each spu thread is allocated.  The address of the control
// block is passed as a startup parameter.  The control block is then DMA'd by
// the spu program to obtain control variables.  At the end of the spu program 
// the control block is DMA'd back to PPU memory to return the results of the 
// spu program.
spu_rand_CB_t  spu_cb[NUMSPUS]  __attribute__ ((aligned(128)));

// Number of simulations for each SPE to run.  This will be an output parameter
// from mc_rand_sb_distribute_sims and an input parameter to
// mc_rand_sb_create_spu_seeds.
int nSims[NUMSPUS];

// Sobol seed for each SPE.  This will be used as the output parameter from 
// mc_rand_sb_create_spu_seeds.  
unsigned int sobolSeed[NUMSPUS];

typedef struct thread_args {
	spe_context_ptr_t spe;
	void * argp;
	void * envp;
} thread_args_t;

void *spe_thread(void *arg){
	unsigned int runflags = 0;
	unsigned int entry = SPE_DEFAULT_ENTRY;
	thread_args_t * targ = (thread_args_t *)arg;
    // run SPE context
	if(spe_context_run(targ->spe, &entry, runflags, targ->argp, targ->envp, NULL) <0 ) {
		printf("** spe_context_run error, errno = %d\n", errno);
	  exit(1);
	}
	pthread_exit(NULL);
}


// ******************************************************************************************************
int get_args(int argc, char** argv, UINT32 *u32Dim, UINT32 *u32Runs, UINT32 *u32Iter, UINT32 *u32nSPE)
{
	int i;

	// argument checking and extraction
	if ( (argc!=8) && (argc!=9)){
		printf("\nUSAGE: %s -d <dimensions> -r <runs> -i <iterations> -s <spu count>\n", argv[0]);
		printf(" note: dimension x runs < 7 000  and dimension x runs x iterations < 6 250 000\n");
		printf("if not program terminates \n");
		printf("Example: %s -d 4 -r 1000 -i 1000 -s 8\n", argv[0]);
	  return 1;
	}

	i=atoi(argv[2]);
	if (i < 0) 
	{
		printf("error: dimension must be > 0!\n");
		return 1;
	}
  
	i=atoi(argv[4]);
	if (i < 1) 
	{
		printf("error: runs must be > 0!\n");
		return 1;
	}

	i=atoi(argv[6]);
	if (i < 1)
	{
		printf("error: iterations must be > 0!\n");
		return 1;
	}

        i=atoi(argv[8]);
        if ((i > NUMSPUS) || (i == 0))
        {
  		printf("error: spu count must be > 0 and <= %d\n",NUMSPUS);
        }

    // Start at i = 1 to skip the command name
	for (i = 1; i < argc; i++) {

		// Check for a switch (leading "-")
		if (argv[i][0] == '-') {

			// Use the next character to decide what to do
            switch (argv[i][1]) {
				case 'd':       *u32Dim = atoi(argv[++i]); break;
				case 'r':       *u32Runs = atoi(argv[++i]); break;
				case 'i':       *u32Iter = atoi(argv[++i]); break;
                                case 's':       *u32nSPE = atoi(argv[++i]); break;
				default:        fprintf(stderr, "Unknown argument %s\n", argv[i]); return 1;
			}
		}
	}

	if (((*u32Dim) * (*u32Runs)) > 7000) {
		printf("\n dimension x runs must not larger than 7000 because of local store limitation-> exit \n");
		return 1;
	}

	if (((*u32Dim) * (*u32Runs) * (*u32Iter)) > 6250000) {
		printf("\n dimension x runs x iteration  must not larger than 6 250 000 because of main memory limitation-> exit \n");
		return 1;
	}

	return 0;
}




// ***************************************************************************
int main(int argc, char* argv[])
{
    unsigned int i=0;
	unsigned int u32Iter;
	unsigned int u32Dim;
	unsigned int u32Runs;
        unsigned int u32nSPE;

	pthread_t pts[NUMSPUS];
	spe_context_ptr_t spe[NUMSPUS];
	thread_args_t targ1[NUMSPUS];
	int rc;

        if( 0 != get_args(argc, argv, &u32Dim, &u32Runs, &u32Iter, &u32nSPE)) {
		return 1;
	}

        unsigned int simulations = u32Dim * u32Runs * u32Iter;

	// Initialize an array of sobol seeds, one for each SPE
        if (mc_rand_sb_distribute_sims(simulations,64,u32nSPE,nSims) < 0) {
  	  printf("Sobol sim distribution error\n");
        }	
        if (mc_rand_sb_create_spu_seeds(0,1,u32nSPE,nSims,sobolSeed) < 0) {
          printf("Sobol seed creation error\n");
          return 1;
        }
  
	// loop to set up control block and start thread for each spe
	for(i=0;i<u32nSPE;i++) {
	  //  Set up control block
	spu_cb[i].u32Iter = u32Iter;
	spu_cb[i].u32Dim  = u32Dim;
	spu_cb[i].u32Runs = u32Runs;
	spu_cb[i].u32Index = i;
        spu_cb[i].u32nHits = 0;
        spu_cb[i].u32totalTries = 0;
	rand_sobol_set_CB_1( &(spu_cb[i].sobolCB), 0);
        spu_cb[i].u32Seed = sobolSeed[i];

	  // Create context
	if( (spe[i] = spe_context_create(0, NULL)) == NULL){
		printf("** spe_context_create error, i=%d, errno = %d\n", i, errno);
	    return 1;
	}

	  // Load spe program
	if( spe_program_load(spe[i], &calcPI_spu)!=0){
		printf("** spe_program_load error, i=%d, errno = %d\n", i, errno);
	    return 1;
	}
   
	targ1[i].spe = spe[i];
	targ1[i].argp = (void *)&spu_cb[i];
	targ1[i].envp = NULL;

	  // Create thread
	if ((rc = pthread_create(&pts[i], NULL, spe_thread, (void *)&targ1[i])) != 0) {
		printf ("** pthread_create error, error=%d\n", rc);
	    return 1;
	}
	}  // end of spe starting loop

	// Wait for the SPE threads to terminate.
        for(i=0;i<u32nSPE;++i) {
		if ((rc = pthread_join(pts[i], NULL)) != 0) {
			printf ("** pthread_join error, error=%d)\n", rc);
	    return 1;
		}
        }

                
	// Destroy context
        for(i=0;i<u32nSPE;++i) {
		if ((rc = spe_context_destroy (spe[i])) != 0){
			printf ("**spe_context_destroy error, i=%d, errno=%d \n", i, rc);
	    return 1;
		}	
        }
                
	// Total the results from all spes
       unsigned long nHits = 0;
       unsigned long totalTries = 0;
       for (i=0;i<u32nSPE;i++) {
         nHits += spu_cb[i].u32nHits;
         totalTries += spu_cb[i].u32totalTries;
         printf("SPE %d tries %d, hits %d\n",i,spu_cb[i].u32totalTries,spu_cb[i].u32nHits);
       }
       double ratio = 4 * ((double) nHits / (double) totalTries);

	// print the results
	printf("\nResult from all SPEs is PI = %f\n", ratio);
	for ( i = 0; i < u32nSPE; i++)
	{
	  printf("\nResult from SPE %i is PI = %f",i,spu_cb[i].pi);
	}
	printf("\n");

	return 0;
}
