/* --------------------------------------------------------------  */
/* (C)Copyright 2008                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

// pgm name
//   Monte Carlo simulation spu program.  This program may be run
//   on multiple spus, each spu being initialized from a control 
//   block DMA'd from ppu memory.  The program uses a Sobol
//   quasi-random number generator to create the points to be 
//   tested.  The seed for the random number generator has been
//   chosen by a program running on the ppu that has knowledge
//   of the number of points being sampled, the number of spus
//   being started, and the Sobol dimensions.  The seed is passed
//   in the DMA'd control block.  The number of points in 
//   the circle and the total number of points tested
//   are returned to the ppu in the control block.  A value for 
//   PI calculated just by the points processed in this spu is 
//   also returned in the control block for comparison to the PI
//   value calculated in the PPU for all points from all spus.
 
#include <stdio.h>
#include <spu_mfcio.h>
#include <libmisc.h>
#include <rand_0_to_1_v.h>

#include <math.h>

// NOTE: The following two includes have an order dependency to ensure the Sobol control block is defined.
#include <mc_rand.h>
#include "../piMC.h"

#include "new_rand_functions.h" // contains double versions of SDK random number generators

#define DMA_TAG_0 0
#define DMA_TAG_1 1

#define UINT64 unsigned long long
#define INT32 signed int
#define U32 unsigned int

double RADIUS_SQUARE;
float RADIUS;

int RUNS=0, DIMENSION=0;

spu_rand_CB_t cb __attribute__ ((aligned(128)));		// local copy of control block in main memory

// The appropriate functions for the data type defined in Makefile.defines
// are conditionally compiled in.  These functions test the point and 
// increment the control block counters.

#ifdef FLOAT_T
// *******************************************************************
float mc_float(vector float *a, vector float *b){
	// calculate PI
	int k1, k2;
	float x, y, r, ratio;

	for(k1=0; k1 < RUNS; k1++){
		for (k2 = 0; k2 < 4; k2++){ // for each element in a vector
			x = spu_extract(a[k1], k2);
			y = spu_extract(b[k1], k2);
			r = (x-0.5)*(x-0.5) + (y-0.5)*(y-0.5);
			++cb.u32totalTries;

			// Hit Test
			if( (r < RADIUS_SQUARE) ){ 
				++cb.u32nHits;
			}
		}
	}

	ratio = (float) cb.u32nHits / (float) cb.u32totalTries;

	return ratio;
}
#endif

#ifdef DOUBLE_T
// ***************************************************************************
double mc_double( vector double *a, vector double *b){
	// calculate PI
	int k1, k2;
	double x, y, r, ratio;

	for(k1=0; k1 < RUNS; k1++){
		for (k2 = 0; k2 < 2; k2++){ // for each element in a vector
			x = spu_extract(a[k1], k2);
			y = spu_extract(b[k1], k2);
			r = (x-0.5)*(x-0.5) + (y-0.5)*(y-0.5);
			++cb.u32totalTries;

			// Hit Test
			if( (r < RADIUS_SQUARE) ){ 
				++cb.u32nHits;
			}
		}
	}

	ratio = (float) cb.u32nHits / (float) cb.u32totalTries;

	return ratio;
}
#endif

// ***********************************************************
int main(UINT64 speid, UINT64 argp/*, UINT64 envp*/)
{

	speid = 0;
	// get the contents of the control block from main memory to local store (1)
	// control block is definitly smaller that 4k bytes, hence it can be loaded by one DMA task
	mfc_get(&cb, argp, 128, (UINT32)DMA_TAG_0, 0, 0);

	// set DMA - TAG mask to control DMA status (2)
	mfc_write_tag_mask(1<<DMA_TAG_0);

	// wait until all pending DMA transfers are done (3)
	mfc_read_tag_status_all();

	// variables for Monte Carlo integration
	unsigned int i, j;

        RUNS = cb.u32Runs;
        DIMENSION = cb.u32Dim;
        
#ifdef FLOAT_T
	float z=0, sumS=0;
	// variables for RNGs
	vector float rn_a[cb.u32Runs],    rn_b[cb.u32Runs];
#endif

#ifdef DOUBLE_T
	double z=0, sumS=0;
	// variables for RNGs
	vector double rn_a[cb.u32Runs], rn_b[cb.u32Runs]; 
#endif

	RADIUS_SQUARE = .5*.5;
	sumS = 0;
        
	// initialize the Sobol random number generator
	vector float * pf_vecCacheLine;
	unsigned int u32CacheSize = ((((cb.u32Runs + 76)*cb.u32Dim) + 8) * sizeof(vector float) + 127)&0x0FFFFFF80;
	pf_vecCacheLine = (vector float *)malloc_align(u32CacheSize,7);   // dynamic allocation of 128 = 2^7 byte aligned memory
	vector unsigned char * rand_sb_mem = (vector unsigned char *)((char * )pf_vecCacheLine);
	mc_rand_sb1_init(&(cb.sobolCB), cb.u32Runs, cb.u32Dim, rand_sb_mem, u32CacheSize );
	mc_rand_sb_seed(cb.u32Seed);    // initialize with seed passed from ppu
	printf("Finished Sobol init, seed value = %d\n",cb.u32Seed);

   	for( i = 0; i < (unsigned int)cb.u32Iter; ++i){
		// Create arrays of random numbers 
#ifdef FLOAT_T
		for(j=0;j<(unsigned int)cb.u32Runs;j++){

			rn_a[j] = mc_rand_sb_0_to_1_f4();
			rn_b[j] = mc_rand_sb_0_to_1_f4();

		}

		// MonteCarlo simulation to find PI with RADIUS_SQUARE
		z = mc_float(&rn_a[0], &rn_b[0]);  // 
#endif

#ifdef DOUBLE_T
          
		for(j=0;j<(unsigned int)cb.u32Runs;j++){

			rn_a[j] = mc_rand_sb_0_to_1_d2();
			rn_b[j] = mc_rand_sb_0_to_1_d2();
		}

		// MonteCarlo simulation to find PI with RADIUS_SQUARE
		z = mc_double(&rn_a[0], &rn_b[0]);  // 
#endif


		sumS += z;
	}
printf("Test");
        // calculate a value of PI based on the points processed
	// by this SPU
        cb.pi = (4*sumS)/cb.u32Iter;

        // put results to main memory
        mfc_put(&cb, argp, 128, DMA_TAG_1, 0, 0);
        mfc_write_tag_mask(1<<DMA_TAG_1);
        mfc_read_tag_status_all();

	return 0;
}
